//
// Created by Attila Haluska on 9/8/18.
//

#include "compiler/compiler.h"

#include <stdio.h>

#include "compiler/lexer/token.h"
#include "compiler/parser/ast/ast_node.h"
#include "compiler/parser/parser.h"
#include "data_structures/bytecode_array.h"
#include "vm/value.h"
#include "vm/vm_opcode.h"

#define CONSTANTS_MAX   UINT8_MAX

typedef struct compilation_context_t {
    BytecodeArray* current_bytecode;
    AstNode* current_node;
    bool had_error;
} CompilationContext;

static void compile_ast_node(CompilationContext* context);

static void error(CompilationContext* context, const char* message) {
    context->had_error = true;

    fprintf(stderr, "[line %llu] Compile error", context->current_node->token->line);

    if (context->current_node->token->type == TOKEN_EOF) {
        fprintf(stderr, " at the end of file");
    } else {
        fprintf(stderr, " at '%.*s'", (int) context->current_node->token->length, context->current_node->token->start);
    }

    fprintf(stderr, ": %s\n", message);
}

static void emit_byte(CompilationContext* context, uint8_t byte) {
    bytecode_array_add_bytecode(context->current_bytecode, byte, context->current_node->token->line);
}

static void emit_bytes(CompilationContext* context, uint8_t byte1, uint8_t byte2) {
    emit_byte(context, byte1);
    emit_byte(context, byte2);
}

static uint8_t make_constant(CompilationContext* context, Value* value) {
    uint16_t constant = bytecode_array_add_constant(context->current_bytecode, value);

    if (constant > CONSTANTS_MAX) {
        error(context, "Too many constants in one bytecode");
    }

    return (uint8_t) constant;
}

static void emit_constant(CompilationContext* context, Value* value) {
    emit_bytes(context, OP_CONSTANT, make_constant(context, value));
}

static void compile_integer_node(CompilationContext* context) {
    Value* value = value_new_integer_value(context->current_node->as.integer);
    emit_constant(context, value);
}

static void compile_double_node(CompilationContext* context) {
    Value* value = value_new_double_value(context->current_node->as.double_precision);
    emit_constant(context, value);
}

static void compile_binary_infix_node(CompilationContext* context) {
    AstNode* current_node = context->current_node;

    context->current_node = current_node->as.binary_infix.left;
    compile_ast_node(context);

    context->current_node = current_node->as.binary_infix.right;
    compile_ast_node(context);

    context->current_node = current_node;

    switch (current_node->token->type) {
        case TOKEN_PLUS:
            emit_byte(context, OP_ADD);
            break;
        case TOKEN_MINUS:
            emit_byte(context, OP_SUBTRACT);
            break;
        case TOKEN_STAR:
            emit_byte(context, OP_MULTIPLY);
            break;
        case TOKEN_SLASH:
            emit_byte(context, OP_DIVIDE);
            break;
        default:
            break;
    }
}

static void compile_unary_prefix_node(CompilationContext* context) {
    AstNode* current_node = context->current_node;

    context->current_node = current_node->as.unary_prefix.right;
    compile_ast_node(context);

    context->current_node = current_node;

    switch (context->current_node->token->type) {
        case TOKEN_MINUS:
            emit_byte(context, OP_NEGATE);
            break;
        default:
            break;
    }
}

static void compile_ast_node(CompilationContext* context) {
    switch (context->current_node->type) {
        case NODE_INTEGER:
            compile_integer_node(context);
            break;
        case NODE_DOUBLE:
            compile_double_node(context);
            break;
        case NODE_BINARY_INFIX_OPERATOR:
            compile_binary_infix_node(context);
            break;
        case NODE_UNARY_PREFIX_OPERATOR:
            compile_unary_prefix_node(context);
            break;
    }
}

bool compiler_compile_source(const char* source, BytecodeArray** bytecode) {
    if (bytecode != NULL) {
        AstNode* rootNode;

        if (!parser_parse(source, &rootNode)) {
            return false;
        }

        CompilationContext context = {
                .current_bytecode = *bytecode,
                .current_node = rootNode,
                .had_error = false,
        };

        compile_ast_node(&context);
        emit_byte(&context, OP_RETURN);

        return !context.had_error;
    }

    return true;
}
