//
// Created by Attila Haluska on 9/8/18.
//

#include "vm/vm_functions.h"

#include <stddef.h>
#include <stdarg.h>
#include <stdio.h>

#include "vm/value.h"
#include "vm/vm.h"

uint8_t vm_read_byte(Vm* vm) {
    return *vm->ip++;
}

Value* vm_read_constant(Vm* vm) {
    return vm->code.constants.values[vm_read_byte(vm)];
}

void vm_push(Vm* vm, Value* value) {
    *vm->stack_top = value;
    vm->stack_top++;
}

Value* vm_pop(Vm* vm) {
    *vm->stack_top = NULL;
    vm->stack_top--;
    Value* value = *vm->stack_top;
    return value;
}

Value* vm_peek(Vm* vm, uint8_t distance) {
    return vm->stack_top[-1 - distance];
}

void vm_reset_stack(Vm* vm) {
    vm->stack_top = vm->stack;
}

void vm_runtime_error(Vm* vm, const char* format, ...) {
    va_list args;
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    fputs("\n", stderr);

    size_t instruction = vm->ip - vm->code.code.values - 1;
    fprintf(stderr, "[line %llu] in script\n", vm->code.lines.values[instruction]);

    vm_reset_stack(vm);
}

void vm_binary_operation(Vm* vm, VmBinaryOperation operation, bool* should_return, VmInterpretResult* result) {
    if (IS_NUMBER(vm_peek(vm, 0)) && IS_NUMBER(vm_peek(vm, 1))) {
        Value* b = vm_pop(vm);
        Value* a = vm_pop(vm);

        if (IS_DOUBLE(a) && !(IS_DOUBLE(b))) {
            b = value_new_double_value((long double) AS_INTEGER(b));
        }

        if (IS_DOUBLE(b) && !(IS_DOUBLE(a))) {
            a = value_new_double_value((long double) AS_INTEGER(a));
        }

        Value* new_value = NULL;

        switch (operation) {
            case VM_BIN_OP_ADD:
                if (IS_DOUBLE(a) && IS_DOUBLE(b)) {
                    new_value = value_new_double_value(AS_DOUBLE(a) + AS_DOUBLE(b));
                } else {
                    new_value = value_new_integer_value(AS_INTEGER(a) + AS_INTEGER(b));
                }
                break;
            case VM_BIN_OP_SUBTRACT:
                if (IS_DOUBLE(a) && IS_DOUBLE(b)) {
                    new_value = value_new_double_value(AS_DOUBLE(a) - AS_DOUBLE(b));
                } else {
                    new_value = value_new_integer_value(AS_INTEGER(a) - AS_INTEGER(b));
                }
                break;
            case VM_BIN_OP_MULTIPLY:
                if (IS_DOUBLE(a) && IS_DOUBLE(b)) {
                    new_value = value_new_double_value(AS_DOUBLE(a) * AS_DOUBLE(b));
                } else {
                    new_value = value_new_integer_value(AS_INTEGER(a) * AS_INTEGER(b));
                }
                break;
            case VM_BIN_OP_DIVIDE:
                if (IS_DOUBLE(a) && IS_DOUBLE(b)) {
                    new_value = value_new_double_value(AS_DOUBLE(a) / AS_DOUBLE(b));
                } else {
                    new_value = value_new_integer_value(AS_INTEGER(a) / AS_INTEGER(b));
                }
                break;
            default:
                break;
        }

        vm_push(vm, new_value);

        return;
    }

    *should_return = true;
    *result = VM_INTERPRET_RUNTIME_ERROR;
    vm_runtime_error(vm, "Operands must be numbers");
}
